# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the package_publish workflow."""

import logging
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, ClassVar, override

import debusine.tasks.tags as ttags
from debusine.artifacts.models import CollectionCategory, TaskTypes
from debusine.db.models import Artifact, Collection, WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.server.celery import run_server_task
from debusine.server.workflows import PackagePublishWorkflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    PackagePublishWorkflowData,
    PackagePublishWorkflowDynamicData,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    BaseDynamicTaskData,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    OutputData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


@dataclass
class ExpectedCopy:
    """
    A summary of an expected copy.

    This exists to abbreviate tests, and is only useful for copying
    source/binary artifacts.
    """

    artifact: Artifact
    component: str
    section: str
    priority: str | None = None

    @property
    def source_items(self) -> list[str]:
        return [f"{self.artifact.id}@artifacts"]

    @property
    def variables(self) -> dict[str, str]:
        variables = {"component": self.component, "section": self.section}
        if self.priority is not None:
            variables["priority"] = self.priority
        return variables


class PackagePublishWorkflowTests(WorkflowTestCase[PackagePublishWorkflow]):
    """Unit tests for :py:class:`PackagePublishWorkflow`."""

    SAMPLE_TASK_DATA = {
        "target_suite": (f"bookworm@{CollectionCategory.SUITE}")
    }

    scenario = scenarios.DefaultContext()
    source_artifact: ClassVar[Artifact]
    binary_artifact: ClassVar[Artifact]
    upload_artifact: ClassVar[Artifact]
    bookworm: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.playground.create_group_role(
            cls.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[cls.scenario.user],
        )
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello"
        )
        cls.binary_artifact = (
            cls.playground.create_minimal_binary_package_artifact("hello")
        )
        cls.upload_artifact = (
            cls.playground.create_upload_artifacts(src_name="hello")
        ).upload
        cls.bookworm = cls.create_suite_collection("bookworm")

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["binary_artifacts"] = LookupMultiple(
            (self.upload_artifact.id,)
        )
        self.task_data["source_artifact"] = self.source_artifact.id

    @classmethod
    def create_suite_collection(
        cls, name: str, *, workspace: Workspace | None = None
    ) -> Collection:
        """Create a `debian:suite` collection."""
        return cls.playground.create_collection(
            name=name, category=CollectionCategory.SUITE, workspace=workspace
        )

    def create_package_publish_workflow(
        self, task_data: PackagePublishWorkflowData
    ) -> PackagePublishWorkflow:
        """Create a package_publish workflow."""
        wr = self.playground.create_workflow(
            task_name="package_publish", task_data=task_data
        )
        return self.get_workflow(wr)

    def test_create_orchestrator(self) -> None:
        """A PackagePublishWorkflow can be instantiated."""
        source_artifact = self.source_artifact.pk
        binary_artifacts = (self.binary_artifact.pk,)
        target_suite = self.bookworm.pk

        workflow = self.create_package_publish_workflow(
            PackagePublishWorkflowData(
                source_artifact=source_artifact,
                binary_artifacts=LookupMultiple(binary_artifacts),
                target_suite=target_suite,
            )
        )

        self.assertEqual(workflow.data.source_artifact, source_artifact)
        self.assertEqual(
            workflow.data.binary_artifacts, LookupMultiple(binary_artifacts)
        )
        self.assertEqual(workflow.data.target_suite, target_suite)

    def orchestrate(
        self,
        task_data: PackagePublishWorkflowData,
        source_artifact: Artifact,
        architectures: Sequence[str],
        *,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create and orchestrate a PackagePublishWorkflow."""
        playground = self.playground

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Example pipeline for package publishing."""

            @override
            def populate(self) -> None:
                sbuild, _ = self.work_request_ensure_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact.id),
                        target_distribution="debian:sid",
                        architectures=list(architectures) or ["all"],
                    ),
                    workflow_data=WorkRequestWorkflowData(
                        display_name="sbuild", step="sbuild"
                    ),
                )
                self.orchestrate_child(sbuild)

                package_publish, _ = self.work_request_ensure_child_workflow(
                    task_name="package_publish",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="publish packages", step="package_publish"
                    ),
                )

                if task_data.source_artifact is not None:
                    self.requires_artifact(
                        package_publish, task_data.source_artifact
                    )
                self.requires_artifact(
                    package_publish, task_data.binary_artifacts
                )
                playground.orchestrate_child(self, package_publish)

        template = self.playground.create_workflow_template(
            name="examplepipeline-template",
            task_name="examplepipeline",
            workspace=workspace,
        )
        root = self.playground.create_workflow(task_name="examplepipeline")
        self.assertEqual(root.workspace, template.workspace)
        self.assertTrue(
            self.playground.orchestrate_workflow(root, use_data=False)
        )

        return root

    def assert_copies(
        self,
        work_request: WorkRequest,
        target_collection: Collection,
        expected_copies: Sequence[ExpectedCopy],
    ) -> None:
        """Assert that a task declares the expected copies."""
        self.assertEqual(
            work_request.task_data,
            {
                "copies": [
                    {
                        "source_items": copy.source_items,
                        "target_collection": target_collection.id,
                        "unembargo": False,
                        "replace": False,
                        "variables": copy.variables,
                    }
                    for copy in expected_copies
                ]
            },
        )

    def assert_updates_indexes(
        self, package_publish: WorkRequest, target_suite: Collection
    ) -> None:
        """Assert that the workflow updates indexes after copying."""
        copy_collection_items = package_publish.children.get(
            task_type=TaskTypes.SERVER, task_name="copycollectionitems"
        )
        self.playground.advance_work_request(
            copy_collection_items, assign_new_worker=True
        )
        trigger_update_suites = (
            package_publish.children.workflow_callbacks().get(
                workflow_data_json__step="trigger-update-suites"
            )
        )

        self.assertQuerySetEqual(
            trigger_update_suites.dependencies.all(), [copy_collection_items]
        )

        copy_collection_items_result = run_server_task.apply(
            args=(copy_collection_items.id,)
        )
        self.assertTrue(copy_collection_items_result.get())

        trigger_update_suites.refresh_from_db()
        self.assertEqual(
            trigger_update_suites.status, WorkRequest.Statuses.PENDING
        )

        self.schedule_and_run_workflow_callback(package_publish)
        trigger_update_suites.refresh_from_db()
        self.assertEqual(
            trigger_update_suites.status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(
            trigger_update_suites.result, WorkRequest.Results.SUCCESS
        )

        update_suites = target_suite.workspace.workrequest_set.get(
            task_type=TaskTypes.WORKFLOW, task_name="update_suites"
        )
        self.assertEqual(update_suites.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(
            update_suites.task_data, {"only_suites": [target_suite.name]}
        )

        self.schedule_and_run_workflow(update_suites)
        update_suites.refresh_from_db()
        self.assertEqual(update_suites.status, WorkRequest.Statuses.RUNNING)
        self.assertTrue(update_suites.children.exists())

    def test_target_suite_in_wrong_workspace(self) -> None:
        other_workspace = self.playground.create_workspace(
            name="other", public=True
        )
        self.create_suite_collection("sid", workspace=other_workspace)
        target_suite_lookup = (
            f"//workspace=other/sid@{CollectionCategory.SUITE}"
        )

        with self.assertRaisesRegex(
            TaskConfigError,
            f"{target_suite_lookup} is in workspace {other_workspace},"
            f" not {self.scenario.workspace}",
        ):
            self.create_package_publish_workflow(
                PackagePublishWorkflowData(
                    source_artifact=self.source_artifact.pk,
                    binary_artifacts=LookupMultiple((self.binary_artifact.pk,)),
                    target_suite=target_suite_lookup,
                )
            )

    @preserve_db_task_registry()
    def test_populate_source_and_binary(self) -> None:
        """Test population with both source and binary artifacts."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", architectures={"any", "all"}
        )
        target_suite = self.bookworm
        architectures = ("amd64", "i386", "all")

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
            source_artifact=source_artifact,
            architectures=architectures,
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.BLOCKED)

        sbuild_workflow = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        binaries, _ = self.simulate_sbuild_workflow_completion(
            sbuild_workflow, use_data=False
        )

        package_publish.refresh_from_db()
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(root)

        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        expected = [ExpectedCopy(source_artifact, "main", "misc")]
        for arch in architectures:
            expected += [
                ExpectedCopy(binary_artifact, "main", "misc", "optional")
                for binary_artifact in binaries[arch]
            ]
        self.assert_copies(copy_collection_items, target_suite, expected)
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_source_only(self) -> None:
        """Test population with only a source artifact."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1", architectures={"all"}
        )
        target_suite = self.bookworm

        with self.assertLogs(
            logger="debusine.server.workflows.package_publish",
            level=logging.WARNING,
        ) as log:
            root = self.orchestrate(
                task_data=PackagePublishWorkflowData(
                    source_artifact=source_artifact.id,
                    target_suite=(
                        f"{target_suite.name}@{CollectionCategory.SUITE}"
                    ),
                ),
                source_artifact=source_artifact,
                architectures=[],
            )
        self.assertEqual(
            [record.message for record in log.records],
            [
                f"No component specified when adding hello_1.0-1 to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'main'",
                f"No section specified when adding hello_1.0-1 to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'misc'",
            ],
        )

        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )
        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(source_artifact, "main", "misc")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_source_upload_only(self) -> None:
        """Test population with only a source upload artifact."""
        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="hello",
            src_architectures={"any", "all"},
            binary=False,
            section="editors",
        )
        target_suite = self.bookworm

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
            source_artifact=upload_artifacts.source,
            architectures=[],
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(upload_artifacts.source, "main", "editors")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_missing_binary_fields(self) -> None:
        """Population warns about missing binary component/section/priority."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1", architectures={"all"}
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                package="hello",
                version="1.0-1",
                architecture="all",
            )
        )
        target_suite = self.bookworm

        with self.assertLogs(
            logger="debusine.server.workflows.package_publish",
            level=logging.WARNING,
        ) as log:
            root = self.orchestrate(
                task_data=PackagePublishWorkflowData(
                    binary_artifacts=LookupMultiple((binary_artifact.id,)),
                    target_suite=(
                        f"{target_suite.name}@{CollectionCategory.SUITE}"
                    ),
                ),
                source_artifact=source_artifact,
                architectures=[],
            )
        self.assertEqual(
            [record.message for record in log.records],
            [
                f"No component specified when adding hello_1.0-1_all to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'main'",
                f"No section specified when adding hello_1.0-1_all to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'misc'",
                f"No priority specified when adding hello_1.0-1_all to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to "
                f"'optional'",
            ],
        )

        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )
        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(binary_artifact, "main", "misc", "optional")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_non_main_component(self) -> None:
        """Population handles sections with non-main components."""
        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="hello",
            src_architectures={"any", "all"},
            binary=False,
            section="contrib/games",
        )
        target_suite = self.bookworm

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
            source_artifact=upload_artifacts.source,
            architectures=[],
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(upload_artifacts.source, "contrib", "games")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_honours_suite_variables(self) -> None:
        """Population honours ``suite_variables``."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1", architectures={"all"}
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                package="hello",
                version="1.0-1",
                architecture="all",
            )
        )
        target_suite = self.bookworm

        with self.assertNoLogs(
            logger="debusine.server.workflows.package_publish"
        ):
            root = self.orchestrate(
                task_data=PackagePublishWorkflowData(
                    source_artifact=source_artifact.id,
                    binary_artifacts=LookupMultiple((binary_artifact.id,)),
                    target_suite=(
                        f"{target_suite.name}@{CollectionCategory.SUITE}"
                    ),
                    suite_variables={
                        "component": "non-free",
                        "section": "editors",
                        "priority": "extra",
                    },
                ),
                source_artifact=source_artifact,
                architectures=[],
            )

        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )
        self.assert_copies(
            copy_collection_items,
            target_suite,
            [
                ExpectedCopy(artifact, "non-free", "editors", "extra")
                for artifact in (source_artifact, binary_artifact)
            ],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_no_update_indexes(self) -> None:
        """The workflow can be told not to update indexes."""
        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="hello",
            src_architectures={"any", "all"},
            binary=False,
            section="editors",
        )
        target_suite = self.bookworm

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                update_indexes=False,
            ),
            source_artifact=upload_artifacts.source,
            architectures=[],
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(upload_artifacts.source, "main", "editors")],
        )
        self.assertFalse(
            package_publish.children.workflow_callbacks()
            .filter(workflow_data_json__step="trigger-update-suites")
            .exists()
        )

    @preserve_db_task_registry()
    def test_populate_package_build_logs(self) -> None:
        """Test population with binary artifacts and copying build logs."""
        source_workspace = self.playground.create_workspace(name="source")
        self.playground.create_group_role(
            source_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        target_workspace = self.playground.create_workspace(name="target")
        self.playground.create_group_role(
            target_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE, workspace=target_workspace
        )
        source_workspace.set_inheritance([target_workspace])
        source_suite = self.create_suite_collection(
            "bookworm", workspace=source_workspace
        )
        target_suite = self.create_suite_collection(
            "bookworm", workspace=target_workspace
        )
        source_build_logs_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS,
                workspace=source_workspace,
            )
        )
        target_build_logs_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS,
                workspace=target_workspace,
            )
        )
        binary_names_by_architecture = {
            "amd64": ("hello", "hello-dev"),
            "i386": ("hello", "hello-dev"),
            "all": ("hello-doc",),
        }
        binary_artifacts: list[Artifact] = []
        binary_artifacts_lookup: list[str] = []
        for architecture, binary_names in binary_names_by_architecture.items():
            for binary_name in binary_names:
                binary_artifacts.append(
                    self.playground.create_minimal_binary_package_artifact(
                        package=binary_name,
                        version="1.0-1",
                        architecture=architecture,
                        workspace=source_workspace,
                    )
                )
                source_suite.manager.add_artifact(
                    binary_artifacts[-1],
                    user=self.playground.get_default_user(),
                    variables={
                        "component": "main",
                        "section": "misc",
                        "priority": "optional",
                    },
                )
                binary_artifacts_lookup.append(
                    f"//workspace={source_workspace.name}/bookworm@debian:suite"
                    f"/binary:{binary_name}_{architecture}"
                )

        package_publish = self.playground.create_workflow(
            task_name="package_publish",
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    binary_artifacts_lookup
                ),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                unembargo=True,
            ),
            workspace=target_workspace,
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.playground.orchestrate_workflow(package_publish)

        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        expected_copies: list[dict[str, Any]] = [
            {
                "source_items": [f"{binary_artifact.id}@artifacts"],
                "target_collection": target_suite.id,
                "unembargo": True,
                "replace": False,
                "variables": {
                    "component": "main",
                    "section": "misc",
                    "priority": "optional",
                },
            }
            for binary_artifact in binary_artifacts
        ]
        expected_copies.append(
            {
                "source_items": [
                    {
                        "collection": source_build_logs_collection.id,
                        "lookup_filters": [
                            [
                                "same_work_request",
                                binary_artifacts_lookup,
                            ]
                        ],
                    }
                ],
                "target_collection": target_build_logs_collection.id,
                "unembargo": True,
                "replace": False,
            }
        )
        self.assertEqual(
            copy_collection_items.task_data, {"copies": expected_copies}
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_task_history(self) -> None:
        """Test population with binary artifacts and copying build logs."""
        source_workspace = self.playground.create_workspace(name="source")
        self.playground.create_group_role(
            source_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        target_workspace = self.playground.create_workspace(name="target")
        self.playground.create_group_role(
            target_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE, workspace=target_workspace
        )
        source_workspace.set_inheritance([target_workspace])
        source_suite = self.create_suite_collection(
            "bookworm", workspace=source_workspace
        )
        target_suite = self.create_suite_collection(
            "bookworm", workspace=target_workspace
        )
        source_task_history_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.TASK_HISTORY,
                workspace=source_workspace,
            )
        )
        target_task_history_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.TASK_HISTORY,
                workspace=target_workspace,
            )
        )
        binary_names_by_architecture = {
            "amd64": ("hello", "hello-dev"),
            "i386": ("hello", "hello-dev"),
            "all": ("hello-doc",),
        }
        binary_artifacts: list[Artifact] = []
        binary_artifacts_lookup: list[str] = []
        for architecture, binary_names in binary_names_by_architecture.items():
            for binary_name in binary_names:
                binary_artifacts.append(
                    self.playground.create_minimal_binary_package_artifact(
                        package=binary_name,
                        version="1.0-1",
                        architecture=architecture,
                        workspace=source_workspace,
                    )
                )
                source_suite.manager.add_artifact(
                    binary_artifacts[-1],
                    user=self.playground.get_default_user(),
                    variables={
                        "component": "main",
                        "section": "misc",
                        "priority": "optional",
                    },
                )
                binary_artifacts_lookup.append(
                    f"//workspace={source_workspace.name}/bookworm@debian:suite"
                    f"/binary:{binary_name}_{architecture}"
                )

        package_publish = self.playground.create_workflow(
            task_name="package_publish",
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    binary_artifacts_lookup
                ),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                unembargo=True,
            ),
            workspace=target_workspace,
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.playground.orchestrate_workflow(package_publish)

        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        expected_copies: list[dict[str, Any]] = [
            {
                "source_items": [f"{binary_artifact.id}@artifacts"],
                "target_collection": target_suite.id,
                "unembargo": True,
                "replace": False,
                "variables": {
                    "component": "main",
                    "section": "misc",
                    "priority": "optional",
                },
            }
            for binary_artifact in binary_artifacts
        ]
        expected_copies.append(
            {
                "source_items": [
                    {
                        "collection": source_task_history_collection.id,
                        "lookup_filters": [
                            ["same_workflow", binary_artifacts_lookup]
                        ],
                    }
                ],
                "target_collection": target_task_history_collection.id,
                "unembargo": True,
                "replace": False,
            }
        )
        self.assertEqual(
            copy_collection_items.task_data, {"copies": expected_copies}
        )
        self.assert_updates_indexes(package_publish, target_suite)

    def test_populate_fails_with_inconsistent_binary_artifact_workspaces(
        self,
    ) -> None:
        workspaces = {
            name: self.playground.create_workspace(name=name)
            for name in ("source1", "source2", "target")
        }
        for workspace in workspaces.values():
            self.playground.create_group_role(
                workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
            )
        self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE, workspace=workspaces["target"]
        )
        workspaces["source1"].set_inheritance([workspaces["target"]])
        workspaces["source2"].set_inheritance([workspaces["target"]])
        suites = {
            name: self.create_suite_collection("bookworm", workspace=workspace)
            for name, workspace in workspaces.items()
        }
        binary_artifacts: list[Artifact] = []
        binary_artifacts_lookup: list[str] = []
        for architecture, suite in (
            ("amd64", suites["source1"]),
            ("i386", suites["source2"]),
        ):
            binary_artifacts.append(
                self.playground.create_minimal_binary_package_artifact(
                    package="hello",
                    version="1.0-1",
                    architecture=architecture,
                    workspace=suite.workspace,
                )
            )
            suite.manager.add_artifact(
                binary_artifacts[-1],
                user=self.playground.get_default_user(),
                variables={
                    "component": "main",
                    "section": "misc",
                    "priority": "optional",
                },
            )
            binary_artifacts_lookup.append(
                f"//workspace={suite.workspace.name}/bookworm@debian:suite"
                f"/binary:hello_{architecture}"
            )

        package_publish = self.playground.create_workflow(
            task_name="package_publish",
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    binary_artifacts_lookup
                ),
                target_suite=(
                    f"{suites['target'].name}@{CollectionCategory.SUITE}"
                ),
                unembargo=True,
            ),
            workspace=workspaces["target"],
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.playground.orchestrate_workflow(package_publish)

        package_publish.refresh_from_db()
        self.assertEqual(package_publish.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(package_publish.result, WorkRequest.Results.ERROR)
        assert isinstance(package_publish.output_data, OutputData)
        assert package_publish.output_data.errors is not None
        self.assertEqual(len(package_publish.output_data.errors), 1)
        self.assertTrue(
            package_publish.output_data.errors[0].message.startswith(
                "Orchestrator failed:"
                " Binary artifacts must all be in the same workspace"
            )
        )
        self.assertEqual(
            package_publish.output_data.errors[0].code, "orchestrator-failed"
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.GROUP_PREFIX + "debusine:System:owner"}
        )

    def test_compute_dynamic_data_binary_package(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task, strip_task_inputs=True),
            PackagePublishWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello → "
                f"bookworm@{CollectionCategory.SUITE}",
                source_artifact_id=self.source_artifact.id,
                target_suite_id=self.bookworm.id,
                binary_artifacts_ids=[self.upload_artifact.id],
            ),
        )

    def test_get_input_artifacts(self) -> None:
        workflow = self.task
        workflow.dynamic_data = self.compute_dynamic_data(workflow)

        assert workflow.data.source_artifact is not None

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[self.upload_artifact.id],
                ),
            ],
        )

    def test_get_input_artifacts_without_source_artifact(self) -> None:
        del self.task_data["source_artifact"]
        workflow = self.task
        workflow.dynamic_data = self.compute_dynamic_data(workflow)

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[self.upload_artifact.id],
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to extract signing input from other artifacts."""

import json
import re
import shlex
import shutil
import subprocess
import tempfile
from pathlib import Path, PurePath
from typing import Any, override

from debusine.artifacts import SigningInputArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    WorkRequestResults,
    get_architectures,
    get_binary_package_name,
    get_binary_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import BaseTaskWithExecutor, RunCommandTask
from debusine.tasks.inputs import (
    EnvironmentInput,
    SingleInput,
    Stage,
    UploadArtifactsInput,
)
from debusine.tasks.models import (
    ExtractForSigningData,
    ExtractForSigningDynamicData,
)

# https://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-source
_re_package_name = re.compile(r"^[a-z0-9][a-z0-9+.-]+$")


class ExtractError(Exception):
    """An error occurred while extracting signing input."""


class ExtractForSigning(
    RunCommandTask[ExtractForSigningData, ExtractForSigningDynamicData],
    BaseTaskWithExecutor[ExtractForSigningData, ExtractForSigningDynamicData],
):
    """Task to extract signing input from other artifacts."""

    TASK_VERSION = 1

    template_artifact = SingleInput(
        field="input.template_artifact",
        categories=[ArtifactCategory.BINARY_PACKAGE],
        stage=Stage.PENDING,
    )
    binary_artifacts = UploadArtifactsInput(
        field="input.binary_artifacts",
        categories=[ArtifactCategory.BINARY_PACKAGE],
    )
    environment = EnvironmentInput()

    @override
    def post_init(self) -> None:
        super().post_init()

        # Set by fetch_input.
        self._template_deb_name: str | None = None
        self._template_path: Path | None = None
        self._binary_artifacts: dict[str, int] | None = None
        self._binary_paths: dict[str, Path] | None = None

        # Set by run.
        self._remote_execute_directory: PurePath | None = None
        self._local_execute_directory: Path | None = None
        self._signing_input_artifacts: (
            dict[str, SigningInputArtifact] | None
        ) = None

    @override
    def _cmdline(self) -> list[str]:
        """Unused abstract method from RunCommandTask."""
        raise NotImplementedError()

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check if the specified worker can run the task."""
        if not super().can_run_on(worker_metadata):
            return False

        executor_available_key = f"executor:{self.backend}:available"
        return bool(worker_metadata.get(executor_available_key, False))

    @override
    def get_subject(self) -> str | None:
        assert isinstance(self.template_artifact.data, DebianBinaryPackage)
        return get_binary_package_name(self.template_artifact.data)

    @override
    def compute_dynamic_data(self) -> ExtractForSigningDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: binary package name of ``template_artifact``
        :parameter_summary: same as subject (architecture)
        """
        template_artifact = self.template_artifact
        assert isinstance(template_artifact.data, DebianBinaryPackage)
        binary_package_name = get_binary_package_name(template_artifact.data)
        version = get_binary_package_version(template_artifact.data)

        binary_artifacts = self.binary_artifacts
        individual_binary_artifact_ids = []
        architectures = set()

        for binary_artifact in binary_artifacts.artifacts:
            individual_binary_artifact_ids.append(binary_artifact.artifact_id)
            assert isinstance(binary_artifact.data, DebianBinaryPackage)
            architectures |= get_architectures(binary_artifact.data)

        architectures_str = ", ".join(sorted(architectures))

        return ExtractForSigningDynamicData(
            environment_id=self.environment.artifact_id,
            input_template_artifact_id=template_artifact.artifact_id,
            input_binary_artifacts_ids=individual_binary_artifact_ids,
            subject=self.get_subject(),
            parameter_summary=f"{binary_package_name}_{version} "
            f"({architectures_str})",
        )

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the required artifacts."""
        assert self.dynamic_data is not None

        (template_destination := destination / "template").mkdir()
        artifact = self.fetch_artifact(
            self.dynamic_data.input_template_artifact_id, template_destination
        )
        if artifact.category != ArtifactCategory.BINARY_PACKAGE:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"Expected template_artifact to be of category "
                    f"{ArtifactCategory.BINARY_PACKAGE}; got "
                    f"{artifact.category}"
                ],
            )
            return False
        self._template_deb_name = artifact.data["deb_fields"]["Package"]
        # Checked by BinaryPackage validator.
        assert len(artifact.files) == 1
        template_file = list(artifact.files)[0]
        if not template_file.endswith(".deb"):
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"Expected template_artifact file name to match *.deb; "
                    f"got {template_file}"
                ],
            )
            return False
        self._template_path = template_destination / template_file

        (binary_destination := destination / "binary").mkdir()
        self._binary_artifacts = {}
        self._binary_paths = {}
        for artifact_id in self.dynamic_data.input_binary_artifacts_ids:
            artifact = self.fetch_artifact(artifact_id, binary_destination)
            # Checked by build_dynamic_data.
            assert artifact.category == ArtifactCategory.BINARY_PACKAGE
            binary_deb_name = artifact.data["deb_fields"]["Package"]
            self._binary_artifacts[binary_deb_name] = artifact_id
            # Checked by BinaryPackage validator.
            assert len(artifact.files) == 1
            binary_file = list(artifact.files)[0]
            self._binary_paths[binary_deb_name] = (
                binary_destination / binary_file
            )

        return True

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """Configure task: create and start an executor instance."""
        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        return True

    def _run_cmd_or_raise(
        self, cmd: list[str], execute_directory: PurePath
    ) -> None:
        self.logger.info("Executing: %s", shlex.join(cmd))
        returncode = self.run_cmd(cmd, Path(execute_directory))
        self.logger.info("%s exited with code %s", cmd[0], returncode)
        if returncode != 0:
            raise ExtractError(f"{cmd[0]} exited with code {returncode}")

    def _pull_from_executor(self, source: PurePath, target: Path) -> None:
        """Pull a subdirectory from the executor, preserving timestamps."""
        assert self.executor_instance is not None
        assert self._remote_execute_directory is not None
        assert self._local_execute_directory is not None

        # Guard against accidentally passing
        # {local,remote}_execute_directory directly.
        assert source != self._remote_execute_directory
        assert target != self._local_execute_directory

        source_tar_path = PurePath(f"{source}.tar")
        target_tar_path = Path(f"{target}.tar")
        self._run_cmd_or_raise(
            ["tar", "-cf", str(source_tar_path), source.name], source.parent
        )
        self.executor_instance.file_pull(source_tar_path, target_tar_path)
        subprocess.run(
            ["tar", "--one-top-level", "-xf", target_tar_path],
            cwd=target.parent,
        )

    def _extract_binary(self, deb: PurePath, target: PurePath) -> None:
        """Extract a binary package."""
        self._run_cmd_or_raise(
            ["dpkg-deb", "-x", str(deb), str(target)],
            target.parent,
        )

    def _read_manifest(self) -> dict[str, Any]:
        """Read the files.json manifest from the template artifact."""
        assert self.executor_instance is not None
        assert self._template_deb_name is not None
        assert self._template_path is not None
        assert self._remote_execute_directory is not None
        assert self._local_execute_directory is not None

        remote_template_path = (
            self._remote_execute_directory / self._template_deb_name
        )
        local_files_json_path = self._local_execute_directory / "files.json"

        self._extract_binary(self._template_path, remote_template_path)
        self.executor_instance.file_pull(
            remote_template_path
            / "usr"
            / "share"
            / "code-signing"
            / self._template_deb_name
            / "files.json",
            local_files_json_path,
        )
        manifest = json.loads(local_files_json_path.read_text())
        # For mypy.  If the manifest isn't a JSON object, we'll get a
        # run-time error almost immediately afterwards anyway.
        assert isinstance(manifest, dict)
        return manifest

    def _make_signing_input_artifact(
        self, package: str, metadata: dict[str, Any]
    ) -> SigningInputArtifact:
        """
        Check a package's metadata and make a signing-input artifact.

        :raises ExtractError: if the package's metadata is invalid.
        """
        assert self.executor_instance is not None
        assert self._binary_paths is not None
        assert self._remote_execute_directory is not None
        assert self._local_execute_directory is not None

        if _re_package_name.match(package) is None:
            raise ExtractError(f"'{package}' is not a valid package name")

        remote_binary_path = self._remote_execute_directory / package
        self._extract_binary(self._binary_paths[package], remote_binary_path)

        local_binary_path = self._local_execute_directory / package
        self._pull_from_executor(remote_binary_path, local_binary_path)

        signing_input_paths: list[Path] = []
        for file in metadata["files"]:
            file_path = PurePath(file["file"])
            if ".." in file_path.parts:
                raise ExtractError(
                    f"File name '{file_path}' may not contain '..' segments"
                )
            if file_path.is_absolute():
                raise ExtractError(
                    f"File name '{file_path}' may not be absolute"
                )
            if not (
                (local_binary_path / file_path)
                .resolve()
                .is_relative_to(local_binary_path)
            ):
                raise ExtractError(
                    f"File name '{file_path}' may not traverse symlinks to "
                    f"outside the package"
                )
            signing_input_paths.append(local_binary_path / file_path)

        return SigningInputArtifact.create(
            signing_input_paths,
            local_binary_path.parent,
            trusted_certs=metadata.get("trusted_certs"),
            binary_package_name=package,
        )

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Do the main extraction work."""
        assert self.executor_instance is not None
        assert self._binary_paths is not None

        self._remote_execute_directory = PurePath(execute_directory)
        self._local_execute_directory = Path(
            tempfile.mkdtemp(prefix="debusine-extract-for-signing-")
        )

        manifest = self._read_manifest()
        self._signing_input_artifacts = {
            package: self._make_signing_input_artifact(package, metadata)
            for package, metadata in manifest["packages"].items()
        }

        return WorkRequestResults.SUCCESS

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,
    ) -> None:
        """Upload artifacts for the task."""
        assert self.work_request_id is not None
        assert self.debusine is not None

        if execution_result == WorkRequestResults.SUCCESS:
            assert self.dynamic_data is not None
            assert self._binary_artifacts is not None
            assert self._signing_input_artifacts is not None

            for package, artifact in self._signing_input_artifacts.items():
                uploaded_artifact = self.debusine.upload_artifact(
                    artifact,
                    workspace=self.workspace_name,
                    work_request=self.work_request_id,
                )
                for related_to_id in (
                    self.dynamic_data.input_template_artifact_id,
                    self._binary_artifacts[package],
                ):
                    self.debusine.relation_create(
                        uploaded_artifact.id,
                        related_to_id,
                        RelationType.RELATES_TO,
                    )

    @override
    def cleanup(self) -> None:
        """Clean up after running the task."""
        if self._local_execute_directory is not None:
            shutil.rmtree(self._local_execute_directory)
        super().cleanup()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests related to worker pool views."""

from typing import ClassVar, override

from django.urls import reverse

from debusine.db.models import WorkerPool
from debusine.db.playground import scenarios
from debusine.test.django import TestCase, TestResponseType
from debusine.web.views.tests.utils import ViewTestMixin


class WorkerPoolsListViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkerPoolsListView`."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_template_with_workerpools(self) -> None:
        """Test template contains expected information."""
        pool1 = self.playground.create_worker_pool(name="first")
        pool2 = self.playground.create_worker_pool(
            name="second", provider_account=pool1.provider_account
        )

        self.playground.create_worker(worker_pool=pool1)
        self.playground.create_worker(worker_pool=pool2)
        self.playground.create_worker(worker_pool=pool2)

        response = self.client.get(reverse("worker-pools:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker_pool-list']")

        # Name
        self.assertTextContentEqual(table.tbody.tr[0].td[0], "first")
        self.assertEqual(
            table.tbody.tr[0].td[0].a.get("href"), pool1.get_absolute_url()
        )
        self.assertTextContentEqual(table.tbody.tr[1].td[0], "second")
        self.assertEqual(
            table.tbody.tr[1].td[0].a.get("href"), pool2.get_absolute_url()
        )
        # Enabled
        self.assertTextContentEqual(table.tbody.tr[0].td[1], "yes")
        self.assertTextContentEqual(table.tbody.tr[1].td[1], "yes")
        # Instance-wide
        self.assertTextContentEqual(table.tbody.tr[0].td[2], "yes")
        self.assertTextContentEqual(table.tbody.tr[1].td[2], "yes")
        # Ephemeral
        self.assertTextContentEqual(table.tbody.tr[0].td[3], "no")
        self.assertTextContentEqual(table.tbody.tr[1].td[3], "no")
        # Workers
        self.assertTextContentEqual(table.tbody.tr[0].td[4], "1")
        self.assertTextContentEqual(table.tbody.tr[1].td[4], "2")

        self.assertFalse(tree.xpath("//p[@id='worker_pool-list-empty']"))

    def test_template_no_workerpools(self) -> None:
        """Test template renders "No worker pools registered in this..."."""
        response = self.client.get(reverse("worker-pools:list"))
        tree = self.assertResponseHTML(response)
        p = self.assertHasElement(tree, "//p[@id='worker_pool-list-empty']")
        self.assertTextContentEqual(
            p, "No worker pools registered in this debusine instance."
        )


class WorkerPoolsListViewOrderingTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkerPoolListView` ordering."""

    scenario = scenarios.DefaultContext(set_current=True)
    pool1: ClassVar[WorkerPool]
    pool2: ClassVar[WorkerPool]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.pool1 = cls.playground.create_worker_pool(name="first")
        cls.pool2 = cls.playground.create_worker_pool(
            name="second", provider_account=cls.pool1.provider_account
        )

        cls.playground.create_worker(worker_pool=cls.pool1)
        cls.playground.create_worker(worker_pool=cls.pool1)
        cls.playground.create_worker(worker_pool=cls.pool2)

    def assertTableEntries(
        self, response: TestResponseType, expected: list[str]
    ) -> None:
        """Check that the workers in the table match the given list of names."""
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker_pool-list']")
        actual = [
            self.get_node_text_normalized(tr.td[0]) for tr in table.tbody.tr
        ]
        self.assertEqual(actual, expected)

    def test_name_ascending(self) -> None:
        """Sort by name ascending."""
        response = self.client.get(
            reverse("worker-pools:list"), {"order": "name"}
        )
        self.assertTableEntries(response, ["first", "second"])

    def test_name_descending(self) -> None:
        """Sort by name descending."""
        response = self.client.get(
            reverse("worker-pools:list"), {"order": "name.desc"}
        )
        self.assertTableEntries(response, ["second", "first"])

    def test_workers_ascending(self) -> None:
        """Sort by workers ascending."""
        response = self.client.get(
            reverse("worker-pools:list"), {"order": "workers"}
        )
        self.assertTableEntries(response, ["second", "first"])

    def test_workers_descending(self) -> None:
        """Sort by workers descending."""
        response = self.client.get(
            reverse("worker-pools:list"), {"order": "workers.desc"}
        )
        self.assertTableEntries(response, ["first", "second"])


class WorkerPoolDetailViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkerPoolDetailView`."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_default(self) -> None:
        """Test a GET with an default playground pool."""
        wp = self.playground.create_worker_pool()
        self.client.force_login(self.scenario.user)
        response = self.client.get(wp.get_absolute_url())
        tree = self.assertResponseHTML(response)

        el = self.assertHasElement(tree, "//tr[@id='row-name']")
        self.assertTextContentEqual(el.td[0], wp.name)

        el = self.assertHasElement(tree, "//tr[@id='row-provider_type']")
        self.assertTextContentEqual(
            el.td[0], wp.provider_account.data["provider_type"]
        )

        el = self.assertHasElement(tree, "//tr[@id='row-enabled']")
        self.assertTextContentEqual(el.td[0], "yes" if wp.enabled else "no")

        el = self.assertHasElement(tree, "//tr[@id='row-instance_wide']")
        self.assertTextContentEqual(
            el.td[0], "yes" if wp.instance_wide else "no"
        )

        el = self.assertHasElement(tree, "//tr[@id='row-ephemeral']")
        self.assertTextContentEqual(el.td[0], "yes" if wp.ephemeral else "no")

        el = self.assertHasElement(tree, "//tr[@id='row-registered_at']")
        self.assertDatetimeContentEqual(el, wp.registered_at, anywhere=True)

        el = self.assertHasElement(tree, "//tr[@id='row-architectures']")
        self.assertTextContentEqual(el.td[0], ", ".join(wp.architectures))

        el = self.assertHasElement(tree, "//tr[@id='row-tags']")
        self.assertTextContentEqual(el.td[0], "none")

        el = self.assertHasElement(tree, "//tr[@id='row-specifications']")
        self.assertYAMLContentEqual(el.td[0], wp.specifications)

        el = self.assertHasElement(tree, "//tr[@id='row-limits']")
        self.assertYAMLContentEqual(el.td[0], wp.limits)

        self.assertFalse(tree.xpath("//table[@id='worker-list']"))

    def test_no_architectures(self) -> None:
        """Test with no architectures."""
        wp = self.playground.create_worker_pool()
        wp.architectures = []
        wp.save()
        self.client.force_login(self.scenario.user)
        response = self.client.get(wp.get_absolute_url())
        tree = self.assertResponseHTML(response)
        el = self.assertHasElement(tree, "//tr[@id='row-architectures']")
        self.assertTextContentEqual(el.td[0], "none")

    def test_tags(self) -> None:
        """Test with tags."""
        wp = self.playground.create_worker_pool()
        wp.tags = ["foo", "bar"]
        wp.save()
        self.client.force_login(self.scenario.user)
        response = self.client.get(wp.get_absolute_url())
        tree = self.assertResponseHTML(response)
        el = self.assertHasElement(tree, "//tr[@id='row-tags']")
        self.assertTextContentEqual(el.td[0], "bar, foo")

    def test_workers(self) -> None:
        """Check presentation of worker information."""
        wp = self.playground.create_worker_pool()
        self.playground.create_worker(worker_pool=wp)
        self.playground.create_worker(worker_pool=wp)

        self.client.force_login(self.scenario.user)
        response = self.client.get(wp.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, "//table[@id='worker-list']")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command create_collection."""

import io

import yaml
from django.conf import settings
from django.core.management import CommandError
from django.test import override_settings

from debusine.artifacts.models import CollectionCategory
from debusine.db.models import Collection, default_workspace
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


@override_settings(LANGUAGE_CODE="en-us")
class CreateCollectionCommandTests(TestCase):
    """Tests for the create_collection command."""

    def test_create_collection_from_file(self) -> None:
        """`create_collection` creates a new collection (data in file)."""
        name = "test"
        category = CollectionCategory.SUITE
        data = {"may_reuse_versions": True}
        data_file = self.create_temporary_file(
            contents=yaml.safe_dump(data).encode()
        )
        stdout, stderr, exit_code = call_command(
            "create_collection", name, category, "--data", str(data_file)
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        collection = Collection.objects.get(name=name, category=category)
        self.assertEqual(
            collection.workspace.name, settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        self.assertEqual(collection.data, data)

    def test_create_collection_from_stdin(self) -> None:
        """`create_collection` creates a new collection (data in file)."""
        name = "test"
        category = CollectionCategory.SUITE
        data = {"may_reuse_versions": True}
        stdout, stderr, exit_code = call_command(
            "create_collection",
            name,
            category,
            stdin=io.StringIO(yaml.safe_dump(data)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        collection = Collection.objects.get(name=name, category=category)
        self.assertEqual(
            collection.workspace.name, settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        self.assertEqual(collection.data, data)

    def test_create_collection_empty_data(self) -> None:
        """`create_collection` defaults data to {}."""
        name = "test"
        category = CollectionCategory.ENVIRONMENTS
        stdout, stderr, exit_code = call_command(
            "create_collection", name, category, stdin=io.StringIO()
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        collection = Collection.objects.get(name=name, category=category)
        self.assertEqual(collection.data, {})

    def test_create_collection_different_workspace(self) -> None:
        """`create_collection` can use a non-default workspace."""
        name = "test"
        category = CollectionCategory.ENVIRONMENTS
        workspace_name = "test-workspace"
        workspace = self.playground.create_workspace(name=workspace_name)
        stdout, stderr, exit_code = call_command(
            "create_collection",
            name,
            category,
            "--workspace",
            workspace_name,
            stdin=io.StringIO(),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        collection = Collection.objects.get(name=name, category=category)
        self.assertEqual(collection.workspace, workspace)

    def test_create_collection_different_scope(self) -> None:
        """`create_collection` can use a workspace in a different scope."""
        name = "test"
        category = CollectionCategory.ENVIRONMENTS
        scope_name = "test-scope"
        scope = self.playground.get_or_create_scope(name=scope_name)
        workspace = self.playground.create_workspace(
            scope=scope, name=settings.DEBUSINE_DEFAULT_WORKSPACE
        )
        stdout, stderr, exit_code = call_command(
            "create_collection",
            name,
            category,
            "--workspace",
            f"{scope_name}/{settings.DEBUSINE_DEFAULT_WORKSPACE}",
            stdin=io.StringIO(),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        collection = Collection.objects.get(name=name, category=category)
        self.assertEqual(collection.workspace, workspace)

    def test_create_collection_invalid_data_yaml(self) -> None:
        """`create_collection` returns error: cannot parse YAML data."""
        with self.assertRaisesRegex(
            CommandError, r"^Error parsing YAML:"
        ) as exc:
            call_command(
                "create_collection",
                "test",
                CollectionCategory.ENVIRONMENTS,
                stdin=io.StringIO(":"),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_create_collection_nonexistent_workspace(self) -> None:
        """`create_collection` returns error: workspace not found."""
        with self.assertRaisesRegex(
            CommandError,
            rf"^Workspace 'nonexistent' not found in scope "
            rf"'{settings.DEBUSINE_DEFAULT_SCOPE}'$",
        ) as exc:
            call_command(
                "create_collection",
                "test",
                CollectionCategory.ENVIRONMENTS,
                "--workspace",
                "nonexistent",
                stdin=io.StringIO(),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_create_collection_duplicated_name_and_category(self) -> None:
        """`create_collection` returns error: duplicated name and category."""
        Collection.objects.create(
            name="test",
            category=CollectionCategory.SUITE,
            workspace=default_workspace(),
            data={},
        )
        with self.assertRaisesRegex(
            CommandError,
            r"^Error creating collection: "
            r"Collection with this Name, Category and Workspace already "
            r"exists\.$",
        ) as exc:
            call_command(
                "create_collection",
                "test",
                CollectionCategory.SUITE,
                stdin=io.StringIO(yaml.safe_dump({"may_reuse_versions": True})),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_create_collection_invalid_name(self) -> None:
        """`create_collection` returns error: name is not valid."""
        with self.assertRaisesRegex(
            CommandError,
            r"^Error creating collection: "
            r"'_test' is not a valid collection name",
        ):
            call_command(
                "create_collection",
                "_test",
                CollectionCategory.TEST,
                stdin=io.StringIO(),
            )

        with self.assertRaisesRegex(
            CommandError,
            r"^Error creating collection: "
            r"'a/b' is not a valid collection name",
        ):
            call_command(
                "create_collection",
                "a/b",
                CollectionCategory.TEST,
                stdin=io.StringIO(),
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""API for provisioning and managing cloud workers."""

from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING, cast, override

from debusine.assets import APTAuthenticationData, CloudProvidersType

if TYPE_CHECKING:
    from debusine.db.models import Worker, WorkerPool

from debusine.server.worker_pools.models import WorkerPoolSpecifications

providers: dict[CloudProvidersType, type["WorkerPoolInterface"]] = {}


class WorkerPoolInterface(ABC):
    """Interface for all cloud compute providers."""

    worker_pool: "WorkerPool"
    specifications: WorkerPoolSpecifications

    def __init__(self, worker_pool: "WorkerPool") -> None:
        """Initialize WorkerPoolInterface."""
        self.worker_pool = worker_pool
        self.specifications = worker_pool.specifications_model

    @override
    def __init_subclass__(
        cls, cloud_provider: CloudProvidersType, **kwargs: Any
    ) -> None:
        """Register an instance of WorkerPoolInterface by cloud_provider."""
        super().__init_subclass__(**kwargs)
        providers[cloud_provider] = cls

    @abstractmethod
    def launch_worker(self, worker: "Worker") -> None:
        """
        Launch a cloud worker using worker.

        The implementation is expected to set worker.instance_created_at to the
        current timestamp, and store state in worker.worker_pool_data.

        This method is idempotent, repeatedly calling it on the same worker
        should not create additional cloud instances or restart existing ones.
        """

    def _terminate_existing_instances(
        self, worker_pool_data: dict[str, Any] | None
    ) -> None:
        """Terminate existing instances referenced by ``worker_pool_data``."""

    def terminate_worker(self, worker: "Worker") -> None:
        """
        Terminate the cloud worker specified.

        The implementation sets worker.instance_created_at to None and
        clears worker.worker_pool_data if it was able to successfully
        initiate instance termination.

        This method is idempotent.
        """
        self._terminate_existing_instances(worker.worker_pool_data)
        worker.worker_pool_data = None
        worker.instance_created_at = None
        tokens = (worker.token, worker.activation_token)
        worker.token = None
        worker.activation_token = None
        worker.save()
        for token in tokens:
            if token is not None:
                token.delete()

    def terminate_missing(self) -> None:
        """Terminate any missing pool workers."""

    def get_apt_authentication(self) -> APTAuthenticationData | None:
        """Return APT authentication for installing debusine."""
        if self.worker_pool.apt_authentication:
            return cast(
                APTAuthenticationData,
                self.worker_pool.apt_authentication.data_model,
            )
        return None


def provider_interface(worker_pool: "WorkerPool") -> WorkerPoolInterface:
    """Return a WorkerPoolInterface for the specified worker_pool."""
    cloud_provider = worker_pool.provider_account.data["provider_type"]
    return providers[cloud_provider](worker_pool)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_tokens."""

import datetime as dt
from typing import ClassVar, override

from django.utils import timezone

from debusine.db.models import Token, User
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ListTokensCommandTests(TabularOutputTests, TestCase):
    """Tests for the list_tokens command."""

    user: ClassVar[User]
    token: ClassVar[Token]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = User.objects.create_user(
            username="John", email="john@example.com"
        )
        cls.token = cls.playground.create_user_token(user=cls.user)

    def test_list_tokens_no_filtering(self) -> None:
        """list_tokens print the token key."""
        with self.assertPrintsTable() as output:
            call_command('list_tokens')

        self.assertEqual(output.col(0), [self.token.hash])
        self.assertEqual(output.col(1), [self.user.username])
        self.assertEqual(output.col(2), [self.token.created_at.isoformat()])
        self.assertEqual(output.col(3), ["-"])
        self.assertEqual(output.col(4), ["True"])
        self.assertEqual(output.col(5), [self.token.comment])

    def test_list_tokens_expire_at(self) -> None:
        """list_tokens prints `expire_at` if set."""
        self.token.expire_at = timezone.now() + dt.timedelta(hours=1)
        self.token.save()

        with self.assertPrintsTable() as output:
            call_command('list_tokens')

        self.assertEqual(output.col(0), [self.token.hash])
        self.assertEqual(output.col(1), [self.user.username])
        self.assertEqual(output.col(2), [self.token.created_at.isoformat()])
        self.assertEqual(output.col(3), [self.token.expire_at.isoformat()])
        self.assertEqual(output.col(4), ["True"])
        self.assertEqual(output.col(5), [self.token.comment])

    def test_list_tokens_filtered_by_username(self) -> None:
        """list_tokens print the correct filtered owners."""
        self.playground.create_worker_token()
        with self.assertPrintsTable() as output:
            call_command("list_tokens", "--username", "John")
        self.assertEqual(output.col(0), [self.token.hash])

    def test_list_tokens_sorted_by_created_at(self) -> None:
        """Tokens are sorted by created_at."""
        token_2 = self.playground.create_user_token(user=self.user)

        with self.assertPrintsTable() as output:
            call_command('list_tokens')

        # token_1 is displayed before token2 (ordered by created_at)
        self.assertEqual(output.col(0), [self.token.hash, token_2.hash])

        # Make token_1.created_at later than token_2
        self.token.created_at = token_2.created_at + dt.timedelta(minutes=1)
        self.token.save()

        with self.assertPrintsTable() as output:
            call_command('list_tokens')

        # Now token_1 appear after token_2
        self.assertEqual(output.col(0), [token_2.hash, self.token.hash])



# Copyright 2020-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Authentication backend to mark signon-managed authentication."""

from django.contrib.auth.backends import ModelBackend


class SignonAuthBackend(ModelBackend):
    """
    Auth backend for external authentication.

    There is no specific functionality, and it is currently used to mark users
    authenticated via external signon providers.
    """



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the client_utils.py methods."""

import hashlib
import logging
from textwrap import dedent
from typing import override
from unittest import mock

import requests
import responses
from debian.deb822 import Changes, Deb822

from debusine.artifacts.playground import ArtifactPlayground
from debusine.client.client_utils import (
    copy_file,
    dget,
    download_file,
    get_debian_package,
    get_url_contents_sha256sum,
    requests_put_or_connection_error,
)
from debusine.client.exceptions import (
    ClientConnectionError,
    ContentTooLargeError,
    ContentValidationError,
)
from debusine.test import TestCase
from debusine.utils import get_user_agent


class RequestsPutOrConnectionErrorTests(TestCase):
    """Tests for requests_put_or_connection_error."""

    def test_requests_put_raise_exc_without_request(self) -> None:
        """
        requests_put_or_connection_error handles exceptions without a request.

        This is for 100% branch coverage; as far as we know,
        :py:meth:`requests.put` always sets the ``request`` attribute on
        exceptions.
        """
        patcher = mock.patch("debusine.client.client_utils.requests.put")
        mocked = patcher.start()
        mocked.side_effect = requests.exceptions.RequestException(request=None)
        self.addCleanup(patcher.stop)

        with self.assertRaisesRegex(
            ClientConnectionError, r"^Cannot connect\. Error: "
        ):
            requests_put_or_connection_error()

    @responses.activate
    def test_requests_put_or_connection_error_user_agent(self) -> None:
        """requests_put_or_connection_error sends User-Agent header."""
        responses.add(responses.PUT, "http://example.com")
        requests_put_or_connection_error("http://example.com")

        self.assertEqual(
            responses.calls[0].request.headers.get("User-Agent"),
            get_user_agent(),
        )


class DgetDownloadTests(TestCase):
    """Tests for our dget client."""

    @override
    def setUp(self) -> None:
        """Configure fake http responses."""
        super().setUp()
        self.workdir = self.create_temporary_directory()

        self.r_mock = responses.RequestsMock(assert_all_requests_are_fired=True)
        self.r_mock.start()
        self.addCleanup(self.r_mock.stop)
        self.addCleanup(self.r_mock.reset)

        # Swallow log messages
        log = logging.getLogger("debusine.client.client_utils")
        propagate_setting = log.propagate
        log.propagate = False
        self.addCleanup(lambda: setattr(log, "propagate", propagate_setting))

        self.bodies = {
            "foo.deb": b"abc123",
            "foo_1.0.orig.tar.xz": b"test data",
            "foo_1.0-2.debian.tar.xz": b"more test data",
            "foo_1.0-2_source.buildinfo": b"Format: 1.0\nSource: foo\n",
        }
        for filename, body in self.bodies.items():
            self.r_mock.get(f"http://example.com/{filename}", body=body)
        self.set_dsc_response()
        self.set_changes_response()
        self.expected_stats = {
            "foo.deb": {
                "sha256": "6ca13d52ca70c883e0f0bb101e425a89e8624de51db2d239259"
                "3af6a84118090",
                "sha1": "6367c48dd193d56ea7b0baad25b19455e529f5ee",
                "md5": "e99a18c428cb38d5f260853678922e03",
                "size": 6,
            },
        }

    def _set_response(
        self,
        filename: str,
        body: str,
        files: list[str],
        break_hashes: set[str] = set(),
        break_sizes: set[str] = set(),
    ) -> None:
        file_type: type[Deb822]
        if filename.endswith(".changes"):
            file_type = Changes
        else:
            file_type = Deb822
        full_body = (
            body
            + ArtifactPlayground.hash_deb822_files(
                file_type=file_type,
                file_contents={file: self.bodies[file] for file in files},
                break_hashes=break_hashes,
                break_sizes=break_sizes,
            )
        ).encode('utf-8')
        self.bodies[filename] = full_body
        self.r_mock.upsert(
            responses.GET, f"http://example.com/{filename}", body=full_body
        )

    def set_dsc_response(
        self,
        break_hashes: set[str] = set(),
        break_sizes: set[str] = set(),
    ) -> None:
        """Add / replace the .dsc response."""
        self._set_response(
            "foo_1.0-2.dsc",
            dedent(
                """
                Format: 3.0 (quilt)
                Source: foo
                Package-List:
                 foo deb unknown optional arch=any
                """
            ).lstrip(),
            ["foo_1.0.orig.tar.xz", "foo_1.0-2.debian.tar.xz"],
            break_hashes=break_hashes,
            break_sizes=break_sizes,
        )

    def set_changes_response(self) -> None:
        """Add / replace the .changes response."""
        self._set_response(
            "foo_1.0-2_source.changes",
            dedent(
                """
                Format: 1.8
                Source: foo
                """
            ).lstrip(),
            [
                "foo_1.0-2.dsc",
                "foo_1.0-2.debian.tar.xz",
                "foo_1.0-2_source.buildinfo",
            ],
        )

    def test_copy_file_stats(self) -> None:
        src = self.workdir / "src.deb"
        dest = self.workdir / "foo.deb"
        src.write_bytes(self.bodies["foo.deb"])
        stats = copy_file(src, dest)
        self.assertEqual(self.expected_stats["foo.deb"], stats)

    def test_download_file_downloads(self) -> None:
        """Check `download_file` writes the expected file."""
        dest = self.workdir / "foo.deb"
        download_file("http://example.com/foo.deb", dest)
        self.assertTrue(dest.exists())
        self.assertEqual(dest.read_bytes(), b"abc123")

    def test_download_file_stats(self) -> None:
        """Check the return value of `download_file`."""
        dest = self.workdir / "foo.deb"
        stats = download_file("http://example.com/foo.deb", dest)
        self.assertEqual(self.expected_stats["foo.deb"], stats)

    def test_download_file_user_agent(self) -> None:
        """Check `download_file` sends User-Agent header."""
        dest = self.workdir / "foo.deb"
        download_file("http://example.com/foo.deb", dest)

        self.assertEqual(
            self.r_mock.calls[-1].request.headers.get("User-Agent"),
            get_user_agent(),
        )

    def test_download_file_logging(self) -> None:
        """Ensure `download_file` logs its requests."""
        dest = self.workdir / "foo.deb"
        with self.assertLogsContains(
            message="Downloading http://example.com/foo.deb...",
            logger="debusine.client.client_utils",
        ):
            download_file("http://example.com/foo.deb", dest)

    def test_download_file_size(self) -> None:
        """Ensure `download_file` logs the size of files it's downloading."""
        dest = self.workdir / "big.deb"
        self.r_mock.get(
            "http://example.com/big.deb",
            body=b"1234567890" * 1000,
            headers={
                "Content-Length": "10000",
            },
        )
        with self.assertLogs(logger="debusine.client.client_utils") as cm:
            download_file("http://example.com/big.deb", dest)
        self.assertEqual(2, len(cm.records))
        messages = [record.message for record in cm.records]
        self.assertIn("Size: 0.01 MiB", messages)

    def test_dget_file(self) -> None:
        """Test that `dget` can download a .deb."""
        dget("http://example.com/foo.deb", self.workdir)
        dest = self.workdir / "foo.deb"
        self.assertTrue(dest.exists())

    def test_dget_subdir(self) -> None:
        """Ensure `dget` correctly handles sub-directories."""
        dest = self.workdir / "test.deb"
        self.r_mock.get(
            "http://example.com/subdir/test.deb",
            body=b"abc123",
        )
        dget("http://example.com/subdir/test.deb", self.workdir)
        self.assertTrue(dest.exists())

    def test_dget_dsc(self) -> None:
        """Test that `dget` can download a .dsc."""
        dget("http://example.com/foo_1.0-2.dsc", self.workdir)
        self.assertTrue((self.workdir / "foo_1.0-2.dsc").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.debian.tar.xz").exists())
        self.assertTrue((self.workdir / "foo_1.0.orig.tar.xz").exists())

    def test_dget_source_changes(self) -> None:
        """Test that `dget` can download a _source.changes."""
        dget("http://example.com/foo_1.0-2_source.changes", self.workdir)
        self.assertTrue((self.workdir / "foo_1.0-2_source.changes").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.dsc").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.debian.tar.xz").exists())
        # Test that we recurse through the .dsc to find the orig tarball
        self.assertTrue((self.workdir / "foo_1.0.orig.tar.xz").exists())

    def test_dget_only_downloads_files_once(self) -> None:
        """Check that `dget` downloads aren't unnecessarily repeated."""
        with self.assertLogs(logger="debusine.client.client_utils") as cm:
            dget("http://example.com/foo_1.0-2_source.changes", self.workdir)
        records = [
            record.message
            for record in cm.records
            if record.message.startswith("Downloading")
        ]
        self.assertEqual(len(set(records)), len(records))
        self.assertTrue((self.workdir / "foo_1.0-2_source.changes").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.dsc").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.debian.tar.xz").exists())
        self.assertTrue((self.workdir / "foo_1.0.orig.tar.xz").exists())

    def test_dget_garbage_dsc(self) -> None:
        """If we can't parse it, `dget` simply downloads the .dsc."""
        self.r_mock.get(
            "http://example.com/garbage.dsc",
            body=b"\0:" * 100,
        )
        dget("http://example.com/garbage.dsc", self.workdir)
        self.assertTrue((self.workdir / "garbage.dsc").exists())

    def test_dget_garbage_changes(self) -> None:
        """If we can't parse it, `dget` simply downloads the .changes."""
        self.r_mock.get(
            "http://example.com/garbage.changes",
            body=b"\0:" * 100,
        )
        dget("http://example.com/garbage.changes", self.workdir)
        self.assertTrue((self.workdir / "garbage.changes").exists())

    def test_dget_verifies_sha256(self) -> None:
        """`dget` verifies SHA256 hashes."""
        self.set_dsc_response(break_hashes={"sha256"})
        with self.assertRaisesRegex(
            ContentValidationError,
            r"Downloaded file has mis-matching sha256 for "
            r"foo_1\.0\.orig\.tar\.xz \([0-9a-f]{64} != [0-9a-f]{64}\)",
        ):
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)

    def test_dget_verifies_sha1(self) -> None:
        """`dget` verifies SHA1 hashes."""
        self.set_dsc_response(break_hashes={"sha1"})
        with self.assertRaisesRegex(
            ContentValidationError,
            r"Downloaded file has mis-matching sha1 for "
            r"foo_1\.0\.orig\.tar\.xz \([0-9a-f]{40} != [0-9a-f]{40}\)",
        ):
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)

    def test_dget_verifies_md5(self) -> None:
        """`dget` verifies MD5 hashes."""
        self.set_dsc_response(break_hashes={"md5"})
        with self.assertRaisesRegex(
            ContentValidationError,
            r"Downloaded file has mis-matching md5 for "
            r"foo_1\.0\.orig\.tar\.xz \([0-9a-f]{32} != [0-9a-f]{32}\)",
        ):
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)

    def test_dget_verifies_size(self) -> None:
        """`dget` will spot size mismatch between entries in dsc file."""
        self.set_dsc_response(break_sizes={"sha256"})
        with self.assertRaises(ContentValidationError) as cm:
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)
        self.assertEqual(
            "foo_1.0-2.dsc has mis-matching size for foo_1.0.orig.tar.xz "
            "(9 != 10)",
            str(cm.exception),
        )

    def test_dget_verifies_previously_downloaded_hash(self) -> None:
        """`dget` spots files differing between .dsc and .changes files."""
        self.set_dsc_response(break_hashes={"sha256"})
        self.set_changes_response()
        with self.assertRaisesRegex(
            ContentValidationError,
            r"foo_1.0-2\.dsc has mis-matching sha256 for "
            r"foo_1\.0-2\.debian\.tar\.xz \([0-9a-f]{64} != [0-9a-f]{64}\)",
        ):
            dget("http://example.com/foo_1.0-2_source.changes", self.workdir)

    def test_dget_rejects_path_traversal(self) -> None:
        """`dget` will refuse to download a dsc file with / in filenames."""
        self.r_mock.get(
            "http://example.com/traversal.dsc",
            body=dedent(
                """
                Format: 3.0 (quilt)
                Source: foo
                Package-List:
                 foo deb unknown optional arch=any
                Checksums-Sha1:
                 f48dd853820860816c75d54d0f584dc863327a7c 9 foo/bar.tar.xz
                Files:
                 eb733a00c0c9d336e65691a37ab54293 9 foo/bar.tar.xz
                """
            )
            .lstrip()
            .encode('utf-8'),
        )
        with self.assertRaises(ContentValidationError) as cm:
            dget("http://example.com/traversal.dsc", self.workdir)
        self.assertEqual(
            "traversal.dsc contains invalid file name foo/bar.tar.xz",
            str(cm.exception),
        )

    def test_get_debian_package_local_file(self) -> None:
        """`get_debian_package` will return a Path to a local .deb file."""
        source = self.workdir / "foo.deb"
        source.write_bytes(b"test file")
        with get_debian_package(str(source)) as package:
            self.assertEqual(package, source)

    def test_get_debian_package_file_url(self) -> None:
        """`get_debian_package` will return a Path to a file://...deb URL."""
        source = self.workdir / "foo.deb"
        source.write_bytes(b"test file")
        with get_debian_package(f"file://{source}") as package:
            self.assertEqual(package, source)

    def test_get_debian_package_http_url(self) -> None:
        """`get_debian_package` will download a deb by http."""
        with get_debian_package("http://example.com/foo.deb") as package:
            self.assertEqual(package.read_bytes(), b"abc123")

    def test_get_debian_package_ftp_url(self) -> None:
        """`get_debian_package` can't take ftp:// URLs."""
        with (
            self.assertRaisesRegex(ValueError, r"Not a supported URL scheme:"),
            get_debian_package("ftp://example.com/foo.deb"),
        ):
            pass  # pragma: no cover


class GetUrlContentsSha256sumTests(TestCase):
    """Tests for get_url_contents_sha256sum function."""

    @responses.activate
    def test_get_url_contents_sha256sum(self) -> None:
        """get_url_contents_sha256sum function return contents and sha256."""
        url = "https://example.com"
        max_size = 100
        content = b"Test content"
        sha256sum = hashlib.sha256(content).hexdigest()

        responses.add(responses.GET, url, body=content)

        returned_content, returned_sha256sum = get_url_contents_sha256sum(
            url, max_size
        )

        self.assertEqual(returned_content, content)
        self.assertEqual(returned_sha256sum, sha256sum)

    @responses.activate
    def test_get_url_contents_sha256sum_user_agent(self) -> None:
        """get_url_contents_sha256sum function sends User-Agent header."""
        url = "https://example.com"
        responses.add(responses.GET, url, body=b"test")
        get_url_contents_sha256sum(url, 100)

        self.assertEqual(
            responses.calls[0].request.headers.get("User-Agent"),
            get_user_agent(),
        )

    @responses.activate
    def test_get_url_contents_sha256sum_too_large(self) -> None:
        """get_url_contents_sha256sum function raise ContentTooLargeError."""
        url = "https://example.com"
        max_size = 10
        content = b'Test content that is too large for the specified max_size'

        responses.add(responses.GET, url, body=content)

        message = "Content size exceeds maximum allowed size of 10 bytes."
        with self.assertRaisesRegex(ContentTooLargeError, message):
            get_url_contents_sha256sum(url, max_size)

    def test_file_url_without_allow_file(self) -> None:
        """Without allow_file=True, file:// URLs are rejected."""
        path = self.create_temporary_file()

        with self.assertRaises(requests.exceptions.InvalidSchema):
            get_url_contents_sha256sum(f"file://{path}", 10)

    def test_file_url_with_allow_file(self) -> None:
        """With allow_file=True, file:// URLs work."""
        max_size = 100
        content = b"Test content"
        sha256sum = hashlib.sha256(content).hexdigest()
        path = self.create_temporary_file(contents=content)

        returned_content, returned_sha256sum = get_url_contents_sha256sum(
            f"file://{path}", max_size, allow_file=True
        )

        self.assertEqual(returned_content, content)
        self.assertEqual(returned_sha256sum, sha256sum)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the database playground scenarios."""

from typing import Any, override

from django.contrib.auth.hashers import check_password

from debusine.artifacts.models import ArtifactCategory
from debusine.db.context import context
from debusine.db.models import Scope, Workspace
from debusine.db.playground import scenarios
from debusine.db.playground.playground import Playground
from debusine.test.django import TestCase
from debusine.utils import extract_generic_type_arguments


class ScenarioTestBase[S: scenarios.Scenario](TestCase):
    """Base class for scenario tests."""

    scenario_cls: type[S]
    built_scenario: S | None
    scenario_kwargs: dict[str, Any]

    @override
    def setUp(self) -> None:
        """Set up backend for lazily building a scenario."""
        super().setUp()
        self.built_scenario = None
        self.scenario_kwargs = {}

    def build_scenario(self, **kwargs: Any) -> None:
        """Build the scenario for the current test."""
        self.built_scenario = self.scenario_cls(**kwargs)
        self.playground.build_scenario(
            self.built_scenario,
            set_current=kwargs.get("set_current", False),
        )

    @property
    def scenario(self) -> S:
        """Return the scenario to be tested, building it if needed."""
        if self.built_scenario is None:
            self.build_scenario()
        assert self.built_scenario is not None
        return self.built_scenario

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Set scenario_cls from the type argument."""
        super().__init_subclass__(**kwargs)
        [cls.scenario_cls] = extract_generic_type_arguments(
            cls, ScenarioTestBase
        )


class DefaultScopeUserTestBase[DSU: scenarios.DefaultScopeUser](
    ScenarioTestBase[DSU]
):
    """Test DefaultScopeUser-derived scenario."""

    def test_members(self) -> None:
        """Ensure we get the default members."""
        self.assertEqual(self.scenario.scope.name, "debusine")
        self.assertEqual(self.scenario.user.username, "playground")

    def test_set_context(self) -> None:
        """Test setting in context."""
        self.build_scenario(set_current=True)
        self.assertEqual(context.scope, self.scenario.scope)
        self.assertEqual(context.user, self.scenario.user)

    def test_scope_owners(self) -> None:
        """Test scope_owners."""
        group = self.scenario.scope_owners
        self.assertQuerySetEqual(group.users.all(), [])
        self.assertEqual(
            group.scope_roles.get(resource=self.scenario.scope).role,
            Scope.Roles.OWNER,
        )

    def test_assign_role(self) -> None:
        self.assertQuerySetEqual(self.scenario.scope.roles.all(), [])
        with self.scenario.assign_role(self.scenario.scope, Scope.Roles.OWNER):
            assignment = self.scenario.scope.roles.get()
            self.assertEqual(assignment.resource, self.scenario.scope)
            self.assertEqual(assignment.role, Scope.Roles.OWNER)
            self.assertQuerySetEqual(
                assignment.group.users.all(), [self.scenario.user]
            )
        self.assertQuerySetEqual(self.scenario.scope.roles.all(), [])

    def test_build_with_different_defaults(self) -> None:
        other_playground = Playground(
            default_workspace_name="other", default_username="other"
        )
        self.other_scenario = self.scenario_cls()
        other_playground.build_scenario(self.other_scenario)

        def cleanup() -> None:
            del self.other_scenario

        self.addCleanup(cleanup)

        # The scope name is not yet customizable in the playground
        self.assertEqual(self.other_scenario.scope.name, "debusine")
        self.assertEqual(self.other_scenario.user.username, "other")


class DefaultScopeUserTests(
    DefaultScopeUserTestBase[scenarios.DefaultScopeUser]
):
    """Test DefaultScopeUser scenario."""

    @override
    def test_set_context(self) -> None:
        super().test_set_context()
        self.assertIsNone(context.workspace)


class DefaultScopeUserAPITests(
    DefaultScopeUserTestBase[scenarios.DefaultScopeUserAPI]
):
    """Test DefaultScopeUserAPI scenario."""

    @override
    def test_members(self) -> None:
        super().test_members()
        self.assertEqual(self.scenario.user_token.user, self.scenario.user)


class DefaultContextTestBase[DC: scenarios.DefaultContext](
    DefaultScopeUserTestBase[DC]
):
    """Test DefaultContext-derived scenario."""

    @override
    def test_members(self) -> None:
        super().test_members()
        self.assertEqual(self.scenario.workspace.name, "System")
        self.assertTrue(self.scenario.workspace.public)

    @override
    def test_set_context(self) -> None:
        super().test_set_context()
        self.assertEqual(context.workspace, self.scenario.workspace)

    def test_workspace_owners(self) -> None:
        """Test scope_owners."""
        group = self.scenario.workspace_owners
        self.assertQuerySetEqual(group.users.all(), [])
        self.assertEqual(
            group.workspace_roles.get(resource=self.scenario.workspace).role,
            Workspace.Roles.OWNER,
        )

    @override
    def test_build_with_different_defaults(self) -> None:
        super().test_build_with_different_defaults()
        self.assertEqual(self.other_scenario.workspace.name, "other")
        self.assertEqual(
            self.other_scenario.default_task_configuration_collection.workspace,
            self.other_scenario.workspace,
        )


class DefaultContextTests(DefaultContextTestBase[scenarios.DefaultContext]):
    """Test DefaultContext scenario."""


class DefaultContextAPITests(
    DefaultContextTestBase[scenarios.DefaultContextAPI]
):
    """Test DefaultContextAPI scenario."""

    @override
    def test_members(self) -> None:
        super().test_members()
        self.assertEqual(self.scenario.user_token.user, self.scenario.user)


class DefaultSbuildTestBase[DS: scenarios.DefaultSbuild](
    DefaultContextTestBase[DS]
):
    """Test DefaultSbuild-derived scenario."""

    @override
    def test_members(self) -> None:
        super().test_members()
        self.assertEqual(
            self.scenario.source.category, ArtifactCategory.SOURCE_PACKAGE
        )
        self.assertEqual(self.scenario.environment_sid.data["codename"], "sid")
        self.assertEqual(
            self.scenario.environment_sid.category,
            ArtifactCategory.SYSTEM_TARBALL,
        )


class DefaultSbuildTests(DefaultSbuildTestBase[scenarios.DefaultSbuild]):
    """Test DefaultSbuild scenario."""


class DefaultSbuildAPITests(DefaultSbuildTestBase[scenarios.DefaultSbuildAPI]):
    """Test DefaultSbuildAPI scenario."""

    @override
    def test_members(self) -> None:
        super().test_members()
        self.assertEqual(self.scenario.user_token.user, self.scenario.user)


class UIPlaygroundTests(DefaultContextTestBase[scenarios.UIPlayground]):
    """Test UIPlayground scenario."""

    @override
    def test_members(self) -> None:
        """Ensure we get the default members."""
        self.assertTrue(
            check_password("playground", self.scenario.user.password)
        )
        # TODO: the scenario is kind of improvised and could use a dust-off to
        # make it state of the art with how Debusine actually work. When that
        # is done, the resulting expectations can be encoded here


class PackageBuildLogTests(DefaultContextTestBase[scenarios.PackageBuildLog]):
    """Test PackageBuildLog scenario."""

    @override
    def test_members(self) -> None:
        """Ensure we get the default members."""
        self.assertEqual(self.scenario.workspace.name, "System")
        self.assertTrue(self.scenario.workspace.public)


del DefaultScopeUserTestBase
del DefaultContextTestBase
del DefaultSbuildTestBase
